/*
 * TCOS 2.0 specific operation for PKCS #15 initialization
 *
 * Copyright (C) 2002 Olaf Kirch <okir@lst.de>
 * Copyright (C) 2002 g10 Code GmbH
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif
#include <sys/types.h>
#include <string.h>
#include <openssl/bn.h>
#include "opensc.h"
#include "cardctl.h"
#include "pkcs15-init.h"
#include "util.h"

/* Don't spend to much bugs due to buggy code */
static int have_delete_acl (const struct sc_file *file)
{
        const u8 *s  = file->sec_attr;
        int      len = file->sec_attr_len;

        for (; len >= 6; s +=6, len -= 6) {
                if ( *s == 0x60 || *s == 0x61
                     || *s == 0xe4 || *s == 0xe5)
                        return 0;
        }
        error("No delete ACL set - You don't want this!\n");
        return SC_ERROR_INVALID_ARGUMENTS;
}        



/*
 * Update the contents of a PIN file
 */
static int tcos_update_pin(struct sc_card *card, struct pin_info *info)
{
	int		r;
        u8              buf[2+16];

	if ( !info->secret[0] || !*info->secret[0]) {
		error("Invalid PIN value\n");
		return SC_ERROR_INVALID_ARGUMENTS;
        }

        if (strlen(info->secret[0]) < 6) {
		error("PIN too short (TCOS requires at least 6 characters)\n");
		return SC_ERROR_INVALID_ARGUMENTS;
        }
        if (strlen(info->secret[0]) > sizeof(buf)-2) {
		error("PIN too long\n");
		return SC_ERROR_INVALID_ARGUMENTS;
        }

        /* It seems that we need to create the record with the
           max. length we want to allow for a password so that the
           user can later change it to a larger one.  */

        buf[0] = 1;  /* tag  ? */
        buf[1] = sizeof (buf) - 2;
	memset(buf+2, info->pkcs15.pad_char, sizeof (buf)-2);
	memcpy(buf+2, info->secret[0], strlen (info->secret[0]));

	r = sc_append_record(card, buf, sizeof (buf), 0);
	if (r < 0) {
		error("Storing PIN failed: %s\n", sc_strerror(r));
		return r;
        }

	r = sc_card_ctl(card, SC_CARDCTL_TCOS_SETPERM, NULL);
	if (r < 0) {
		error("Set permanent failed: %s\n", sc_strerror(r));
		return r;
        }

        if ( strlen(info->secret[0]) < sizeof(buf)-2) 
                r = sc_change_reference_data (card, SC_AC_CHV,
                                              info->pkcs15.reference,
                                              buf+2, sizeof (buf)-2,
                                              info->secret[0],
                                              strlen(info->secret[0]),
                                              NULL);
        if (r < 0) {
		error("Shortening PIN failed: %s\n", sc_strerror(r));
		return r;
        }

        info->pkcs15_obj.flags |= SC_PKCS15_PIN_FLAG_INITIALIZED;

	return 0;
}

/*
 * Create the PIN file and write the PINs.  Note that we assume that
 * PIN file creation is an atomic operation which is basically true as
 * we have to commit the changes using the SET PERMANENT command.  If
 * we ever want to replace a PIN file we would need the old PIN
 * anyway.  So in this light it is better to allow deletion of the PIN
 * file. */
static int tcos_store_pin(struct sc_profile *profile, struct sc_card *card,
                          struct pin_info *info)
{
	struct sc_file	*pinfile;
	int		r;

	card->ctx->log_errors = 0;
	r = sc_select_file(card, &info->file->file->path, &pinfile);
	card->ctx->log_errors = 1;
	if (r == SC_ERROR_FILE_NOT_FOUND) {
		/* Well, then create the file but first check that a
                   delete ACL has been set */
                if ((r = have_delete_acl (info->file->file)))
                        return r;
                sc_file_dup(&pinfile, info->file->file);
		if ((r = sc_pkcs15init_create_file(profile, pinfile)) >= 0) {
                        /* If messing with the PIN file requires any sort of
                         * authentication, send it to the card now */
                        if ((r = sc_pkcs15init_authenticate(
                                profile, pinfile, SC_AC_OP_UPDATE)) >= 0) {
                                r = tcos_update_pin(card, info);
                        }
                }
	} 


	sc_file_free(pinfile);
	return r;
}

/*
 * Initialize the Application DF and store the PINs
 */
static int tcos_init_app(struct sc_profile *profile, struct sc_card *card)
{
        struct pin_info	*pin1, *pin2;
	int r;

	pin1 = sc_profile_find_pin(profile, "CHV1");
	pin2 = sc_profile_find_pin(profile, "CHV2");
	if (!pin1) {
		fprintf(stderr, "No CHV1 defined\n");
		return 1;
	}

	card->ctx->log_errors = 0;
	r = sc_select_file(card, &profile->df_info.file->path, NULL);
	card->ctx->log_errors = 1;
	if (r == SC_ERROR_FILE_NOT_FOUND) {
		/* Create the application DF */
		if (sc_pkcs15init_create_file(profile, profile->df_info.file))
			return 1;
	} else if (r < 0) {
		fprintf(stderr, "Unable to select application DF: %s\n",
			sc_strerror(r));
		return 1;
	}

	/* Store the PINs */
	if (tcos_store_pin(profile, card, pin1))
		return 1;
	if (pin2 && tcos_store_pin(profile, card, pin2))
                return 1;
	
	return 0;
}


/*
 * Allocate a file
 */
static int tcos_allocate_file(struct sc_profile *profile, struct sc_card *card,
                              unsigned int type, unsigned int num,
                              struct sc_file **out)
{
	struct file_info *templ;
	char name[64], *tag, *desc;

	desc = tag = NULL;
	while (1) {
		switch (type) {
		case SC_PKCS15_TYPE_PRKEY_RSA:
			desc = "RSA private key";
			tag = "private-key";
			break;
		case SC_PKCS15_TYPE_PUBKEY_RSA:
			desc = "RSA public key";
			tag = "public-key";
			break;
		case SC_PKCS15_TYPE_CERT:
			desc = "certificate";
			tag = "certificate";
			break;
		case SC_PKCS15_TYPE_DATA_OBJECT:
			desc = "data object";
			tag = "data";
			break;
		}
		if (tag)
			break;
		/* If this is a specific type such as
		 * SC_PKCS15_TYPE_CERT_FOOBAR, fall back to
		 * the generic class (SC_PKCS15_TYPE_CERT)
		 */
		if (!(type & ~SC_PKCS15_TYPE_CLASS_MASK)) {
			error("File type not supported by card driver");
			return SC_ERROR_INVALID_ARGUMENTS;
		}
		type &= SC_PKCS15_TYPE_CLASS_MASK;
	}

	snprintf(name, sizeof(name)-1, "template-%s-%d", tag, num + 1);
        name[sizeof(name)-1] = 0;
	templ = sc_profile_find_file(profile, name);
        if (!templ) {
		error("Profile doesn't define %s template (%s)", desc, name);
		return SC_ERROR_NOT_SUPPORTED;
	}
	sc_file_dup(out, templ->file);

	return 0;
}


static int store_rsa_record (struct sc_card *card, int recno,
                             int reclen, BIGNUM *a)
{
        u8 bnbuf[256];
        int mpilen, len, r;

        mpilen = reclen - 3;
        len = BN_num_bytes(a);
        if (len > mpilen || reclen > sizeof (bnbuf)) {
                fprintf(stderr, "Key length invalid.\n");
                return SC_ERROR_INVALID_ARGUMENTS;
        }

        memset(bnbuf, 0, sizeof bnbuf);
        bnbuf[0] = recno;
        bnbuf[1] = mpilen+1;
        BN_bn2bin(a, bnbuf+reclen-len);
        r = sc_update_record(card, recno, bnbuf, reclen, SC_RECORD_BY_REC_NR);
        if (r == SC_ERROR_RECORD_NOT_FOUND)
                r = sc_append_record(card, bnbuf, reclen, 0);
        return r;
}

/* Write a private key file.  It is expected that the file has been
   selected, it is empty and not-permanent.  The caller should use
   set-permanent to commit the changes */
static int tcos_store_private_key(struct sc_card *card, RSA *rsa)
{
        int r;
        int reclen = 67; /*FIXME: get the record length from somewhere */

        r = store_rsa_record(card, 1, reclen, rsa->q);
        if (r >= 0)
                r = store_rsa_record(card, 2, reclen, rsa->p);
        if (r >= 0)
                r = store_rsa_record(card, 3, reclen, rsa->iqmp);
        if (r >= 0)
                r = store_rsa_record(card, 4, reclen, rsa->dmq1);
        if (r >= 0)
                r = store_rsa_record(card, 5, reclen, rsa->dmp1);
        if (r >= 0)
                r = store_rsa_record(card, 6, reclen, rsa->e);

        return r;
}


/*
 * Store an RSA private key on the card
 */
static int tcos_store_rsa_key(struct sc_profile *profile,
                              struct sc_card *card,
                              struct sc_key_template *info,  RSA *rsa)
{
        struct sc_file	*file;
	int		r;

        /* We are not able to modify a file because it may already be
           set permanent.  So we delete it and re-create it.  Note
           that the update/append in store_rsa_record does not make
           much sense then. 
        */

	card->ctx->log_errors = 0;
	r = sc_select_file(card, &info->file->path, &file);
	card->ctx->log_errors = 1;
        if (r >= 0 && file->path.len >= 2) {
                struct sc_path path;

                path.type = SC_PATH_TYPE_FILE_ID;
                path.value[0] = file->path.value[file->path.len - 2];
                path.value[1] = file->path.value[file->path.len - 1];
                path.len = 2;
                path.index = 0;

                r = sc_delete_file(card, &path);
                if (r >= 0) 
                        r = SC_ERROR_FILE_NOT_FOUND;
        }

	if (r == SC_ERROR_FILE_NOT_FOUND) {
                /* Alright, let's create the file */
                if ((r = have_delete_acl (info->file)))
                        return r;

                sc_file_dup(&file, info->file);
		if ((r = sc_pkcs15init_create_file(profile, file)) < 0) {
                        fprintf (stderr, "RSA file create problem: %s\n",
                                 sc_strerror(r));
                        goto leave;
                }
        }
        else if (r < 0) {
		fprintf (stderr, "RSA file delete/select problem: %s\n",
                         sc_strerror(r));
		goto leave;
        }

        r = tcos_store_private_key(card, rsa);
	if (r < 0) {
		error("Store private key failed: %s\n", sc_strerror(r));
		goto leave;
        }


	r = sc_card_ctl(card, SC_CARDCTL_TCOS_SETPERM, NULL);
	if (r < 0) {
		error("Set permanent failed: %s\n", sc_strerror(r));
		goto leave;
        }

leave:
	sc_file_free(file);
	return r < 0? r:0;
}


void bind_tcos_operations(struct pkcs15_init_operations *ops)
{
	ops->init_app = tcos_init_app;
	ops->allocate_file = tcos_allocate_file;
	ops->store_rsa = tcos_store_rsa_key;
}
